"""
    Everything explicitly pertaining to NL things come here.

    **Tokenization**
    1. Input
"""
import spacy
from spacy.tokenizer import Tokenizer

# Local Imports
from .utils.goodies import *

def _get_spacy_vocab_(lang: str) -> spacy.vocab.Vocab:
    """
        Returns a spacy Vocab object to be used in tokenizer. Because multilinguality!
        TODO: Hard code languages!
    :param lang: a string of languages
    :return: a spacy vocab object
    """

    if lang.lower() in ['en', 'english']:
        from spacy.lang.en import English as Lang
    elif lang.lower() in ['es', 'español', 'espanyol', 'espanhol', 'espaneol', 'spanish']:
        from spacy.lang.es import Spanish as Lang
    elif lang.lower() in ['de', 'german', 'deutsch']:
        from spacy.lang.de import German as Lang
    else:
        raise UnknownSpacyLang(f"The language {lang} is either not hardcoded in yet, or not available in spaCy. Sorry.")

    nlp = Lang()
    return nlp.vocab


def preproc(data: Dict[str, List[str]], language: Union[spacy.vocab.Vocab, str] = 'en',
            n_threads: int = 1, vocabularize: bool = False) \
        -> (Dict[str, List[List]], Optional[List[str]], Optional[Dict[str, int]]):
    """
        A slightly abstract method to

        |-> tokenize text
        |-> idfy/vocabularize/onehot-indexify text

        **Example**
        tok_txt = preproc({'train': tr_txt, 'test': ts_txt, 'valid': vl_txt}, 'en')
        tok_txt, stoi = preproc( ... , vocabularize = True)

    :param data: A dict of {'any random tag': list/iter of data}
    :param language: A string which decides what language does the next belong to.
    :param n_threads: 1
    :param vocabularize: a flag which if turned on, will also convert text to ID and return relevant mappings
    :return: if vocabularize:
                -> Dict of original keys and values where values are no longer a list of strings but a list of list of ints
                -> A list of unique tokens
                -> A corresponding Dict of {str:int}
            else:
                -> Dict of original keys and values where values are no longer a list of strings but a list of list of strs
    """

    stoi, tok_data = {}, {k: [] for k in data.keys()}

    # Get spacy Vocab
    vocab = _get_spacy_vocab_(language) if type(language) is str else language
    tokenizer = Tokenizer(vocab)

    for key, value in data.items():
        for doc in tokenizer.pipe(value, batch_size=10000, n_threads=5):
            if vocabularize:
                tok_data[key].append([stoi.setdefault(tok.text, len(stoi)) for tok in doc])
            else:
                tok_data[key].append([tok.text for tok in doc])

    if vocabularize:
        return tok_data
    else:
        return tok_data, stoi



"""
    A file which contain simple data iterators.
"""
from .utils.goodies import *


class SimplestSampler:
    """
        Given X and Y matrices (or lists of lists),
            it returns a batch worth of stuff upon __next__
    :return:
    """

    def __init__(self, data, bs: int = 64):

        try:
            assert len(data["x"]) == len(data["y"])
        except AssertionError:

            raise MismatchedDataError(f"Length of x is {len(data['x'])} while of y is {len(data['y'])}")

        self.x = data["x"]
        self.y = data["y"]
        self.n = len(self.x)
        self.bs = bs  # Batch Size

    def __len__(self):
        return self.n // self.bs - (1 if self.n % self.bs else 0)

    def __iter__(self):
        self.i, self.iter = 0, 0
        return self

    def __next__(self):
        """
        @TODO: edge case: Return leftovers.
        :return:
        """
        if self.i + self.bs >= self.n:
            raise StopIteration

        _x, _y = self.x[self.i:self.i + self.bs], self.y[self.i:self.i + self.bs]
        self.i += self.bs

        return _x, _y


class SortishSampler:
    """
        Sample the data so like-sized text appears together.
        Returns an iterator.

        Works well with list
        @TODO: check/optimize for np arrs

        This needs to be re-sorted every iteration.
            Hence the data is duplicated internally.
            Call reset at epoch end to resort it.

            Or you could init a new instance of this :peace:
    """

    def __init__(self, data, _batchsize: int, _seqlen: int = None, _padidx=0):
        """ @TODO: snip everything with seqlen """
        _inputs, _labels = data['x'], data['y']

        try:
            assert len(_inputs) == len(_labels)
        except AssertionError:
            raise MismatchedDataError

        self.bs = _batchsize
        self.padidx = _padidx
        self.x, self.y = self._reshuffle_(**self._sort_(_inputs, _labels))

        self.ptr = 0

    def reset(self):
        self.x, self.y = self._reshuffle_(self.x, self.y)
        self.ptr = 0

    @staticmethod
    def _reshuffle_(x: list, y: list) -> (list, list):
        """
            Shuffles both, things inside a chunk (batch) and batches.
        :param x: list of np arr
        :param y: list of np arr
        :return: (list, list)
        """
        for i in range(len(x)):
            # Shuffle these chunks
            chunk_idx = np.random.permutation(len(x[i]))
            x[i] = x[i][chunk_idx]
            y[i] = y[i][chunk_idx]

        shuffle_idx = np.random.permutation(len(x))
        return [x[i] for i in shuffle_idx], [y[i] for i in shuffle_idx]

    def _sort_(self, x, y):
        idx = sorted(range(len(x)), key=lambda k: -len(x[k]))
        x, y = [x[i] for i in idx], [y[i] for i in idx]

        final_x, final_y = [], []
        for ptr in range(len(x))[::self.bs]:
            # Now take a snippet of x and y based the line below
            chunk_x = x[ptr: ptr + self.bs if ptr + self.bs < len(x) else len(x)]
            chunk_y = y[ptr: ptr + self.bs if ptr + self.bs < len(x) else len(x)]

            # Find snippet's max len
            chunk_len = len(chunk_x[0])

            # Pad and max np arr of this batch
            npx = pad_sequence(chunk_x, chunk_len, self.padidx)
            npy = np.asarray(chunk_y)

            # Shuffle x and y
            chunk_idx = np.random.permutation(len(chunk_x))
            npx, npy = npx[chunk_idx], npy[chunk_idx]

            # Append to final thing
            final_x.append(npx)
            final_y.append(npy)

        # Finally, shuffle final_x,y
        # final_idx = np.random.permutation(len(final_x))
        # return [final_x[i] for i in final_idx], [final_y[i] for i in final_idx]
        return {'x': final_x, 'y': final_y}

    def __iter__(self):
        return self

    def __next__(self):
        """ Iter over self.x, self.y """
        if self.ptr >= len(self.x):
            raise StopIteration
        _x, _y = self.x[self.ptr], self.y[self.ptr]
        self.ptr += 1
        return _x, _y

    def __len__(self):
        return len(self.x)


if __name__ == "__main__":
    X = np.random.randint(0, 100, (200, 4))
    Y = np.random.randint(0, 100, (200, 1))
    bsz = 19
    diter = SimplestSampler(X, Y, bsz)



"""
    A collection of different learning rate iterators, that can be used to change the lr during training.

    The iterators ideally shouldn't be called on their own, but given to the LearningRateScheduler class
        along with optimizers, to better handle per-layer lr changing if needed.

    All lr iterators must have a coherent length.
"""
from .utils.goodies import *
from typing import Union, Type

class LearningRateSchedule:
    """ Empty class signifying that any derived class is a learning rate schedule. Purely cosmetic"""
    pass

class ConstantLR(LearningRateSchedule):
    def __init__(self, highest_lr: float, iterations: int):
        self.lr = highest_lr
        self.len = iterations

    def __iter__(self):
        return self

    def __next__(self):
        return self.lr

    def __len__(self):
        return self.len

    def reset(self):
        pass


class CosineAnnealingLR(LearningRateSchedule):
    """
        # Learning rate iter which decays w.r.t. cosine curve, #cycles times a batch.

        # USAGE:
            ```
                lr_schedule = CosineAnnealingLR(1000,3,0.01,0.001)
                a = [i for i in lr_schedule]
                plt.plot(a)
                plt.show()
            ```
    """

    def __init__(self, iterations: int, cycles: int, highest_lr: float, lowest_lr: float=0.0):
        """

        :param iterations: number of iters. Should be number of batches in one epoch.
        :param cycles: number of time the LR should restart in a batch
        :param highest_lr: The undecayed LR.
        :param lowest_lr: Limit to which the LR should decay.
        """

        self.iterations = iterations
        self.cycles = cycles
        self.lr_h = highest_lr
        self.lr_l = lowest_lr

        iters = self.iterations // self.cycles
        self.arr = (np.cos((np.arange(iters) / iters) * np.pi) / 2 + 0.5) * (self.lr_h - self.lr_l) + self.lr_l
        self.arr = np.tile(self.arr, cycles)

        self.ptr = -1

    def __iter__(self):
        return self

    def __len__(self)->int:
        return self.iterations

    def __next__(self)->float:
        """ Iter over self.arr """
        self.ptr += 1
        if self.ptr >= self.iterations:
            raise StopIteration

        if self.arr.shape[0] <= self.ptr < self.iterations:
            return self.lr_h
        else:
            return self.arr[self.ptr]

    def reset(self)->None:
        """ Reset the ptr """
        self.ptr = -1


class SlantedTriangularLR(LearningRateSchedule):
    """
        # An LR which first grows, and then decays back to zero w.r.t. iter (n_batch*n_epoch)

        # USAGE:
            ```
                lr_schedule = SlantedTriangularLR(1000,0.1,32,0.001)
                a = [i for i in lr_schedule]
                plt.plot(a)
                plt.show()
            ```
    """

    def __init__(self, iterations, cut_frac, ratio, highest_lr):
        """
        :param iterations: should be n_batch*n_epoch
        :param cut_frac: the fraction at which we should stop increasing and start decreasing LR
        :param ratio: another hyperparam deciding the shape of the entire thing.
        :param highest_lr: the undecayed LR.
        :param freeze_mask: is multiplied to the finally yielded values, can let some layers freeze, thus.
        """

        assert cut_frac < 1.0

        self.iter = iterations
        self.cf = cut_frac
        self.ratio = ratio
        self.lr_h = highest_lr
        self.cut = int(self.iter * self.cf)
        self.ptr = -1

    def __iter__(self):
        return self

    def __len__(self):
        return self.iter

    def __next__(self):
        """Iter over self.arr"""
        self.ptr += 1

        if self.ptr >= self.iter:
            raise StopIteration

        if self.ptr < self.cut:
            p = self.ptr / self.cut
        else:
            p = 1 - ((self.ptr - self.cut) / (self.cut * ((1 / self.cf) - 1)))

        return ((1 + float(p) * (self.ratio - 1)) / self.ratio * 1.0) * self.lr_h

    def reset(self):
        pass


class LearningRateScheduler:
    """
        An iterator which manages diff lr schedule iters, based on optimizer param groups.

        For each param_group in the optimzier,
            it will return a learning rate, sampled from given lr iterators.

        ```
            for g in opt.param_groups:
            g['lr'] = 0.1

            opt.param_groups[0]['lr'] = 0.0
            print([g['lr'] for g in opt.param_groups])

            lr = LearningRateScheduler(opt, epochs=10, batches=30)

            lr.get()   # Works for 300 (n_epochs * n_batches) times.
        ```
    """

    def __init__(self, lr_args: dict, lr_iterator: Type[LearningRateSchedule] = ConstantLR, org_lrs: list = None,
                 optimizer: torch.optim = None, freeze_mask: np.ndarray = None):
        """
        :param optimizer: torch.optim thing. Should have appropriate param groups for effective LR scheduling per layer.
        :param lr_args: a bunch of args (dict) intended for the given lr_iterator
        :param lr_iterator: a class reference of the intended lr schedule.
        """

        if optimizer is None and org_lrs is None:
            raise CustomError("No information specified to get highest lr.")
        elif optimizer is None:
            self.org_lrs = org_lrs
        else:
            self.opt = optimizer
            self.org_lrs = [group['lr'] for group in self.opt.param_groups]

        self.mask = np.array(freeze_mask) if freeze_mask is not None else np.ones_like(self.org_lrs)

        self.lr_iters = [lr_iterator(highest_lr=lr, **lr_args) for lr in self.org_lrs]

    def __len__(self):
        return min([len(_iter) for _iter in self.lr_iters])

    def get(self):
        try:
            return [lr_iter.__next__() for lr_iter in self.lr_iters] * self.mask
        except StopIteration:
            raise StopIteration(f"{self.__class__} was called more than the predefined times.")

    def reset(self)->None:
        """ To be called when the lr schedule needs to restart (e.g. at the end of every epoch). """
        for lr_iter in self.lr_iters:
            lr_iter.reset()

    def unfreeze(self):
        """ Find the last frozen layer and unfreeze it"""

        # First check if the mask isn't all ones
        if not (self.mask == 1).all():

            self.mask[np.max(np.where(self.mask == 0))] = 1



from setuptools import setup, find_packages

with open("README.md", "r") as readmefile:
    package_description = readmefile.read()

setup(
    name="my-torch",
    version="0.0.1",
    author="Priyansh Trivedi",
    author_email="mail@priyansh.page",
    description="A transparent boilerplate + bag of tricks to ease my (yours?) (our?) PyTorch dev time.",
    long_description=package_description,
    long_description_content_type="text/markdown",
    url="https://github.com/geraltofrivia/mytorch/",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix"
    ],
    keywords=[
        'deep learning', 'pytorch', 'boilerplate', 'machine learning', 'neural network', 'preprocessing'
    ]
)


import os
import time
import json
import torch
import pickle
import requests
import argparse
import warnings
import traceback
import numpy as np
from pathlib import Path
from collections import namedtuple
from torch.autograd import Function
from typing import List, Dict, Union, Any, Optional

TRACES_FORMAT = {name: i for i, name in enumerate(['train_acc', 'train_loss', 'val_acc'])}


class CustomError(Exception): pass
class MismatchedDataError(Exception): pass
class NotifyAPIKeyNotFoundError(Exception): pass
class NotifyMessageMismatchError(Exception): pass
class ImproperCMDArguments(Exception): pass
class UnknownSpacyLang(ValueError): pass


class BadParameters(Exception):
    def __init___(self, dErrorArguments):
        Exception.__init__(self, "Unexpected value of parameter {0}".format(dErrorArguments))
        self.dErrorArguments = dErrorArguments


class FancyDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.__dict__ = self


class GradReverse(Function):
    """
        Torch function used to invert the sign of gradients (to be used for argmax instead of argmin)
        Usage:
            x = GradReverse.apply(x) where x is a tensor with grads.
    """
    @staticmethod
    def forward(ctx, x):
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output.neg()


def pad_sequence(matrix_seq: Union[list, np.array], max_length: int = -1, padidx: Any = 0) -> np.array:
    """
        Pads 2D data.

        + Works with list of list as well as numpy matrix
        - Does not work with string inputs.

    :param matrix_seq: a matrix of list
    :param max_length: desired pad len (if not provided, will pad with max length in matrix_seq)
    :param padidx: the id with which to pad the data (could be anything)
    :return: a padded np array
    """

    if max_length < 0:
        max_length = max([len(x) for x in matrix_seq])

    pad_matrix = np.zeros((len(matrix_seq), max_length)) + padidx
    for i, arr in enumerate(matrix_seq):
        pad_matrix[i, :min(max_length, len(arr))] = arr[:min(max_length, len(arr))]

    return pad_matrix


def update_lr(opt: torch.optim, lrs: Union[int, float, list, np.array]) -> Union[int, float, list, np.array]:
    """ Updates lr of the opt. Give it one num for uniform update. Arr otherwise """

    if type(lrs) in [float, int]:
        for grp in opt.param_groups:
            grp['lr'] = lrs
    else:

        # Check for lens
        assert len(opt.param_groups) == len(lrs), f"Mismatch b/w param group ({len(opt.param_groups)}) " \
                                                  f"and lr list sizes ({len(lrs)})."
        for grp, lr in zip(opt.param_groups, lrs):
            grp['lr'] = lr

    return lrs


def make_opt(model, opt_fn: torch.optim, lr: float = 0.001):
    """
        Based on model.layers it creates diff param groups in opt.
    """
    assert hasattr(model, 'layers'), "The model does not have a layers attribute. Check TODO-URL for a how-to"
    return opt_fn([{'params': l.parameters(), 'lr': lr} for l in model.layers])


def default_eval(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
        Expects a batch of input

        :param y_pred: tensor of shape (b, nc)
        :param y_true: tensor of shape (b, 1)
    """
    return torch.mean((torch.argmax(y_pred, dim=1) == y_true).float())


class Timer:
    """ Simple block which can be called as a context, to know the time of a block. """
    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.interval = self.end - self.start


class Counter(dict):
    """ Assumes a list of data (words?), and counts their occurrence. """

    def __init__(self, data):
        super().__init__()

        for datum in data:
            self[datum] = self.get(datum, 0) + 1

    def most_common(self, n):
        return [(x, self[x]) for x in sorted(self.keys(), key=lambda w: -self[w])[:n]]

    def sorted(self, data=None):
        """
            If data given, then sort that and return, if not, sort self.
        :param data: dict
        :return:
        """
        if not data:
            data = self
        return [(x, data[x]) for x in sorted(data.keys(), key=lambda w: -data[w])]

    def cropped_with_freq(self, f):
        return sorted({tok: freq for tok, freq in self.items() if freq > f})


tosave = namedtuple('ObjectsToSave', 'fname obj')

def mt_save_dir(parentdir: Path, _newdir: bool = False):
    """
            Function which returns the last filled/or newest unfilled folder in a particular dict.
            Eg.1
                parentdir is empty dir
                    -> mkdir 0
                    -> cd 0
                    -> return parentdir/0

            Eg.2
                ls savedir -> 0, 1, 2, ... 9, 10, 11
                    -> mkdir 12 && cd 12 (if newdir is True) else 11
                    -> return parentdir/11 (or 12)

            ** Usage **
            Get a path using this function like so:
                parentdir = Path('runs')
                savedir = save_dir(parentdir, _newdir=True)

        :param parentdir: pathlib.Path object of the parent directory
        :param _newdir: bool flag to save in the last dir or make a new one
        :return: None
    """

    # Check if the dir exits
    try:
        assert parentdir.exists(), f'{parentdir} does not exist. Making it'
    except AssertionError:
        parentdir.mkdir(parents=False)

    # List all folders within, and convert them to ints
    existing = sorted([int(x) for x in os.listdir(parentdir) if x.isdigit()], reverse=True)

    if not existing:
        # If no subfolder exists
        parentdir = parentdir / '0'
        parentdir.mkdir()
    elif _newdir:
        # If there are subfolders and we want to make a new dir
        parentdir = parentdir / str(existing[0] + 1)
        parentdir.mkdir()
    else:
        # There are other folders and we dont wanna make a new folder
        parentdir = parentdir / str(existing[0])

    return parentdir

def mt_save(savedir: Path, message: str = None, message_fname: str = None, torch_stuff: list = None, pickle_stuff: list = None,
            numpy_stuff: list = None, json_stuff: list = None):
    """

        Saves bunch of diff stuff in a particular dict.

        NOTE: all the stuff to save should also have an accompanying filename, and so we use tosave named tuple defined above as
            tosave = namedtuple('ObjectsToSave','fname obj')

        ** Usage **
        # say `encoder` is torch module, and `traces` is a python obj (dont care what)
        parentdir = Path('runs')
        savedir = save_dir(parentdir, _newdir=True)
        save(
                savedir,
                torch_stuff = [tosave(fname='model.torch', obj=encoder)],
                pickle_stuff = [tosave('traces.pkl', traces)]
            )


    :param savedir: pathlib.Path object of the parent directory
    :param message: a message to be saved in the folder alongwith (as text)
    :param torch_stuff: list of tosave tuples to be saved with torch.save functions
    :param pickle_stuff: list of tosave tuples to be saved with pickle.dump
    :param numpy_stuff: list of tosave tuples to be saved with numpy.save
    :param json_stuff: list of tosave tuples to be saved with json.dump
    :return: None
    """

    assert savedir.is_dir(), f'{savedir} is not a directory!'

    # Commence saving shit!
    if message:
        with open(savedir / 'message.txt' if message_fname is None else savedir / message_fname, 'w+') as f:
            f.write(message)

    for data in torch_stuff or ():
        try:
            torch.save(data.obj, savedir / data.fname)
        except:
            traceback.print_exc()

    for data in pickle_stuff or ():
        try:
            pickle.dump(data.obj, open(savedir / data.fname, 'wb+'))
        except:
            traceback.print_exc()

    for data in numpy_stuff or ():
        try:
            np.save(savedir / data.fname, data.obj)
        except:
            traceback.print_exc()

    for data in json_stuff or ():
        try:
            json.dump(data.obj, open(savedir / data.fname, 'w+'))
        except:
            traceback.print_exc()


def str2bool(v)->bool:
    """
        Function (copied from -https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse )

    :param v:
    :return:
    """
    if v.__class__ is bool:
        return v
    elif v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def compute_mask(t: Union[torch.Tensor, np.array], padding_idx=0):
    """
    compute mask on given tensor t
    :param t: either a tensor or a nparry
    :param padding_idx: the ID used to represented padded data
    :return: a mask of the same shape as t
    """
    if type(t) is np.ndarray:
        mask = np.not_equal(t, padding_idx)*1.0
    else:
        mask = torch.ne(t, padding_idx).float()
    return mask


def send_notification(data: dict, key: str = None, message_template: str = None, title: str = None) -> None:
    """
        Code which can send notification to a phone (using push.techulus.com).
        It tries to find api key from disk if not given as arg.

        And data is put on a template of message to send it to a phone

    :param data: a dictionary containing (i) epoch count (int); (ii) accuracy (float) and (iii) save directory (str)
    :param key: a str containing an API key (optional).
    :param message_template: a str with placeholders to put in data .
        Eg. 'Model saved at %(directory)s, after %(epoch)d epochs, having %(accuracy)f accuracy'
    :param title: a str which is the title of the notification
    :return: None
    """

    if not key:
        try:
            key = open('./push-techulus-key', 'r').read()
        except FileNotFoundError:
            raise NotifyAPIKeyNotFoundError("Couldn't find key to send a notification with")

    if not title:
        title = 'Bring out the champagne, your model is trained!'

    if not message_template:
        message_template = \
            'A model trained for %(epoch)d epochs, which achieved' \
            ' a %(accuracy)s percent accuracy, is now stored at %(directory)s'

    # Check if data fits the template
    try:
        message = message_template % data
    except (KeyError, TypeError) as e:
        raise NotifyMessageMismatchError(f"The data with keys {list(data.keys())} does not fit the message template.")

    url = "https://push.techulus.com/api/v1/notify"
    payload = json.dumps({"title": title, "body": message})

    headers = {
        'Content-Type': "application/json",
        'x-api-key': key,
    }

    try:
        _ = requests.request("POST", url, data=payload, headers=headers)
        print("Successfully delivered a notification on your cellphone. Cheers!")
    except:     # @TODO: Figure out which exceptions to catch
        traceback.print_exc()
        warnings.warn("Couldn't deliver notifications. Apologies. Report the traceback as an issue on Github, please?")


# Transparent, and simple argument parsing FTW!
def convert_nicely(arg, possible_types=(bool, float, int, str)):
    """ Try and see what sticks. Possible types can be changed. """
    for data_type in possible_types:
        try:

            if data_type is bool:
                # Hard code this shit
                if arg in ['T', 'True', 'true']: return True
                if arg in ['F', 'False', 'false']: return False
                raise ValueError
            else:
                proper_arg = data_type(arg)
                return proper_arg
        except ValueError:
            continue
    # Here, i.e. no data type really stuck
    warnings.warn(f"None of the possible datatypes matched for {arg}. Returning as-is")
    return arg


def parse_args(raw_args: List[str], compulsory: List[str] = (), compulsory_msg: str = "",
               types: Dict[str, type] = None, discard_unspecified: bool = False):
    """
        I don't like argparse.
        Don't like specifying a complex two liner for each every config flag/macro.

        If you maintain a dict of default arguments, and want to just overwrite it based on command args,
        call this function, specify some stuff like

    :param raw_args: unparsed sys.argv[1:]
    :param compulsory: if some flags must be there
    :param compulsory_msg: what if some compulsory flags weren't there
    :param types: a dict of confignm: type(configvl)
    :param discard_unspecified: flag so that if something doesn't appear in config it is not returned.
    :return:
    """

    parsed = {}

    while True:

        try:                                        # Get next value
            nm = raw_args.pop(0)
        except IndexError:                          # We emptied the list
            break

        # Get value
        try:
            vl = raw_args.pop(0)
        except IndexError:
            raise ImproperCMDArguments(f"A value was expected for {nm} parameter. Not found.")

        # Get type of value
        if types:
            try:
                parsed[nm] = types[nm](vl)
            except ValueError:
                raise ImproperCMDArguments(f"The value for {nm}: {vl} can not take the type {types[nm]}! ")
            except KeyError:                    # This name was not included in the types dict
                if not discard_unspecified:     # Add it nonetheless
                    parsed[nm] = convert_nicely(vl)
                else:                           # Discard it.
                    continue
        else:
            parsed[nm] = convert_nicely(vl)

    # Check if all the compulsory things are in here.
    for key in compulsory:
        try:
            assert key in parsed
        except AssertionError:
            raise ImproperCMDArguments(compulsory_msg + f"Found keys include {[k for k in parsed.keys()]}")

    # Finally check if something unwanted persists here
    return parsed
